/*
** pork_acct.c - account management.
** Copyright (C) 2002 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <string.h>
#include <sys/types.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_misc.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_aim.h>
#include <pork_conf.h>

extern struct screen screen;

static int pork_acct_find_helper(void *l, void *r) {
	char *left = l;
	struct aim_acct *acct = r;

	return (strcasecmp(left, acct->username));
}

inline dlist_t *pork_acct_find(char *user) {
	return (dlist_find(screen.acct_list, user, pork_acct_find_helper));
}

inline struct aim_priv *pork_acct_get_priv(char *user) {
	dlist_t *node;
	struct aim_acct *acct;

	node = dlist_find(screen.acct_list, user, pork_acct_find_helper);
	if (node == NULL || node->data == NULL)
		return (NULL);
	acct = node->data;

	return (acct->priv);
}

struct aim_acct *pork_acct_add(char *user) {
	dlist_t *node;
	struct aim_acct *acct;
	char nname[NUSER_LEN];
	struct aim_priv *priv;

	normalize(user, nname, sizeof(nname));
	node = pork_acct_find(nname);
	if (node != NULL)
		return (NULL);

	priv = pork_priv_init(nname);
	if (priv == NULL)
		return (NULL);

	acct = xcalloc(1, sizeof(*acct));
	acct->username = xstrdup(nname);
	acct->priv = priv;
	acct->session = &priv->aim_session;

	screen.acct_list = dlist_add_head(screen.acct_list, acct);
	return (acct);
}

int pork_acct_del(char *user) {
	dlist_t *node;
	struct aim_acct *acct;
	char nname[NUSER_LEN];

	normalize(user, nname, sizeof(nname));
	node = pork_acct_find(nname);
	if (node == NULL)
		return (-1);

	acct = node->data;
	screen.acct_list = dlist_remove(screen.acct_list, node);

	node = screen.window_list;
	if (node != NULL) {
		do {
			struct imwindow *win = node->data;

			if (win->priv == acct->priv) {
				if (win->blist_visible)
					imwindow_blist_hide(win);

				win->priv->ref_count--;
				win->priv = &screen.null_priv;
				win->priv->ref_count++;
			}

			node = node->next;
		} while (node != screen.window_list);
	}

	/* This must always be the case. */
	if (acct->priv->ref_count == 0)
		pork_priv_free(acct->priv);

	free(acct->username);
	free(acct);

	return (0);
}

void pork_acct_del_all(void) {
	dlist_t *cur = screen.acct_list;

	while (cur != NULL) {
		struct aim_acct *acct = cur->data;
		dlist_t *next = cur->next;

		pork_acct_del(acct->username);
		cur = next;
	}
}

void acct_update_blist_format(void *data, void *nothing __notused) {
	struct aim_acct *acct = data;
	struct aim_priv *priv = acct->priv;

	if (priv->blist == NULL)
		return;

	blist_changed_format(priv->blist);
}

void pork_acct_update(void *data, void *nothing __notused) {
	struct aim_acct *acct = data;
	struct aim_priv *priv = acct->priv;

	if (priv->connected) {
		time_t time_now = time(NULL);

		/*
		** We've got to keep track of buddies' idle time
		** ourselves; the aim server reports an idle time
		** for them only when there's a change (i.e. they
		** go idle, come back from idle or their client reports
		** a new idle time).  This routine will run once a minute
		** and increment the idle time of buddies where appropriate.
		*/

		if (time_now >= priv->last_update + 60) {
			aim_keepalive(priv);
			buddy_update_idle(priv);
			priv->last_update = time_now;
			aim_cleansnacs(&priv->aim_session, 59);
		}

		if (priv->buddy_pref != NULL && priv->buddy_pref->report_idle &&
			!priv->marked_idle && opt_get_bool(OPT_REPORT_IDLE))
		{
			time_t time_diff = time_now - priv->last_input;
			int idle_after = opt_get_int(OPT_IDLE_AFTER);

			if (idle_after > 0 && time_diff >= 60 * idle_after) {
				screen_status_msg("Marking you idle after %d minutes of inactivity",
					time_diff / 60);
				aim_set_idle(priv, time_diff);
			}
		}
	}
}

/*
** Initialize an "aim_priv" structure for the
** account owned by the specified user.
** Each user account gets its own priv structure,
** which contains account specific data.
*/

struct aim_priv *pork_priv_init(const char *user) {
	struct aim_priv *priv;

	priv = xcalloc(1, sizeof(*priv));
	priv->username = xstrdup(user);

	if (buddy_init(priv) != 0) {
		free(priv->username);
		free(priv);
		return (NULL);
	}

	read_user_config(priv, priv->username);
	blist_init(priv);

	aim_setup(priv);
	time(&priv->last_input);
	return (priv);
}

/*
** Free everything that needs freeing in the
** given "aim_priv" structure.
*/

void pork_priv_free(struct aim_priv *priv) {
	aim_kill_all_conn(priv);

	save_user_config(priv, priv->username);

	blist_destroy(priv->blist);

	buddy_destroy(priv);

	if (priv->passwd != NULL) {
		memset(priv->passwd, 0, strlen(priv->passwd));
		free(priv->passwd);
	}

	free(priv->blist);
	free(priv->username);
	free(priv->proxy);
	free(priv->proxy_user);
	free(priv->proxy_passwd);
	free(priv->server);
	free(priv->profile);
	free(priv->away_msg);

	free(priv);
}
